@@ -18,6 +18,7 @@
 #include <QMimeData>
 #include <QMutex>
 
+#include <KConfigGroup>
 #include <KLocalizedString>
 #include <KSharedConfig>
 #if KRUNNER_BUILD_DEPRECATED_SINCE(5, 65)
@@ -486,15 +487,15 @@ AbstractRunnerPrivate::~AbstractRunnerPrivate()
 
 void AbstractRunnerPrivate::init()
 {
+    minLetterCount = runnerDescription.value(QStringLiteral("X-Plasma-Runner-Min-Letter-Count"), 0);
     if (runnerDescription.isValid()) {
         const auto rawData = runnerDescription.rawData();
-        minLetterCount = rawData.value(QStringLiteral("X-Plasma-Runner-Min-Letter-Count")).toInt();
         if (rawData.contains(QStringLiteral("X-Plasma-Runner-Match-Regex"))) {
             matchRegex = QRegularExpression(rawData.value(QStringLiteral("X-Plasma-Runner-Match-Regex")).toString());
             hasMatchRegex = matchRegex.isValid() && !matchRegex.pattern().isEmpty();
         }
-        hasUniqueResults = rawData.value(QStringLiteral("X-Plasma-Runner-Unique-Results")).toBool();
-        hasWeakResults = rawData.value(QStringLiteral("X-Plasma-Runner-Weak-Results")).toBool();
+        hasUniqueResults = runnerDescription.value(QStringLiteral("X-Plasma-Runner-Unique-Results"), false);
+        hasWeakResults = runnerDescription.value(QStringLiteral("X-Plasma-Runner-Weak-Results"), false);
     }
 }
 
