@@ -16,24 +16,27 @@
 #include <QDBusPendingReply>
 #include <QIcon>
 #include <QMutexLocker>
+#include <qobjectdefs.h>
 
 #include "dbusutils_p.h"
 #include "krunner_debug.h"
 
 #define IFACE_NAME "org.kde.krunner1"
 
-DBusRunner::DBusRunner(const KPluginMetaData &pluginMetaData, QObject *parent)
-    : Plasma::AbstractRunner(pluginMetaData, parent)
-    , m_mutex(QMutex::NonRecursive)
+DBusRunner::DBusRunner(QObject *parent, const KPluginMetaData &pluginMetaData, const QVariantList &args)
+    : Plasma::AbstractRunner(parent, pluginMetaData, args)
 {
     qDBusRegisterMetaType<RemoteMatch>();
     qDBusRegisterMetaType<RemoteMatches>();
     qDBusRegisterMetaType<RemoteAction>();
     qDBusRegisterMetaType<RemoteActions>();
     qDBusRegisterMetaType<RemoteImage>();
+    qRegisterMetaType<QMap<QString, RemoteActions>>("QMap<QString, RemoteActions>");
 
     QString requestedServiceName = pluginMetaData.value(QStringLiteral("X-Plasma-DBusRunner-Service"));
     m_path = pluginMetaData.value(QStringLiteral("X-Plasma-DBusRunner-Path"));
+    m_hasUniqueResults = pluginMetaData.value(QStringLiteral("X-Plasma-Runner-Unique-Results"), false);
+    m_callLifecycleMethods = pluginMetaData.value(QStringLiteral("X-Plasma-API")) == QLatin1String("DBus2");
 
     if (requestedServiceName.isEmpty() || m_path.isEmpty()) {
         qCWarning(KRUNNER) << "Invalid entry:" << pluginMetaData.name();
@@ -76,15 +79,16 @@ DBusRunner::DBusRunner(const KPluginMetaData &pluginMetaData, QObject *parent)
         // don't check when not wildcarded, as it could be used with DBus-activation
         m_matchingServices << requestedServiceName;
     }
-    if (pluginMetaData.rawData().value(QStringLiteral("X-Plasma-Request-Actions-Once")).toVariant().toBool()) {
-        requestActions();
-    } else {
-        connect(this, &AbstractRunner::prepare, this, &DBusRunner::requestActions);
-    }
+
+    m_requestActionsOnce = pluginMetaData.value(QStringLiteral("X-Plasma-Request-Actions-Once"), false);
+    connect(this, &AbstractRunner::teardown, this, &DBusRunner::teardown);
 
     // Load the runner syntaxes
-    const QStringList syntaxes = pluginMetaData.rawData().value(QStringLiteral("X-Plasma-Runner-Syntaxes")).toVariant().toStringList();
-    const QStringList syntaxDescriptions = pluginMetaData.rawData().value(QStringLiteral("X-Plasma-Runner-Syntax-Descriptions")).toVariant().toStringList();
+    const QJsonValue syntaxesJson = pluginMetaData.rawData().value(QStringLiteral("X-Plasma-Runner-Syntaxes"));
+    const QStringList syntaxes = syntaxesJson.isArray() ? syntaxesJson.toVariant().toStringList() : syntaxesJson.toString().split(QLatin1Char(','), Qt::SkipEmptyParts);
+    const QJsonValue syntaxDescriptionsJson = pluginMetaData.rawData().value(QStringLiteral("X-Plasma-Runner-Syntax-Descriptions"));
+    const QStringList syntaxDescriptions =
+        syntaxDescriptionsJson.isArray() ? syntaxDescriptionsJson.toVariant().toStringList() : syntaxDescriptionsJson.toString().split(QLatin1Char(','), Qt::SkipEmptyParts);
     const int descriptionCount = syntaxDescriptions.count();
     for (int i = 0; i < syntaxes.count(); ++i) {
         const QString &query = syntaxes.at(i);
@@ -95,34 +99,103 @@ DBusRunner::DBusRunner(const KPluginMetaData &pluginMetaData, QObject *parent)
 
 DBusRunner::~DBusRunner() = default;
 
-void DBusRunner::requestActions()
+void DBusRunner::reloadConfiguration()
 {
-    clearActions();
-    m_actions.clear();
+    // If we have already loaded a config, but the runner is told to reload it's config
+    if (m_callLifecycleMethods) {
+        suspendMatching(true);
+        requestConfig();
+    }
+}
 
+void DBusRunner::createQActionsFromRemoteActions(const QMap<QString, RemoteActions> &remoteActions)
+{
+    for (auto it = remoteActions.begin(), end = remoteActions.end(); it != end; it++) {
+        const QString service = it.key();
+        const RemoteActions actions = it.value();
+        auto &serviceActions = m_actions[service];
+        qDeleteAll(serviceActions);
+        serviceActions.clear();
+        for (const RemoteAction &action : actions) {
+            auto a = new QAction(QIcon::fromTheme(action.iconName), action.text, this);
+            a->setData(action.id);
+            serviceActions.append(a);
+        }
+    }
+}
+
+void DBusRunner::teardown()
+{
+    if (m_matchWasCalled) {
+        for (const QString &service : std::as_const(m_matchingServices)) {
+            auto method = QDBusMessage::createMethodCall(service, m_path, QStringLiteral(IFACE_NAME), QStringLiteral("Teardown"));
+            QDBusConnection::sessionBus().asyncCall(method);
+        }
+    }
+    m_actionsForSessionRequested = false;
+    m_matchWasCalled = false;
+}
+
+QMap<QString, RemoteActions> DBusRunner::requestActions()
+{
     // in the multi-services case, register separate actions from each plugin in case they happen to be somehow different
     // then match together in matchForAction()
+    QMap<QString, RemoteActions> returnedActions;
+    for (const QString &service : std::as_const(m_matchingServices)) {
+        // if we only want to request the actions once and have done so we want to skip the service
+        // but in case it got newly loaded we need to request the actions, BUG: 435350
+        if (m_requestActionsOnce) {
+            if (m_requestedActionServices.contains(service)) {
+                continue;
+            } else {
+                m_requestedActionServices << service;
+            }
+        }
 
-    for (const QString &service : qAsConst(m_matchingServices)) {
         auto getActionsMethod = QDBusMessage::createMethodCall(service, m_path, QStringLiteral(IFACE_NAME), QStringLiteral("Actions"));
-        QDBusPendingReply<RemoteActions> reply = QDBusConnection::sessionBus().asyncCall(getActionsMethod);
-
-        auto watcher = new QDBusPendingCallWatcher(reply);
-        connect(watcher, &QDBusPendingCallWatcher::finished, this, [this, watcher, service]() {
-            watcher->deleteLater();
-            QDBusReply<RemoteActions> reply = *watcher;
-            if (!reply.isValid()) {
-                qCDebug(KRUNNER) << "Error requestion actions; calling" << service << " :" << reply.error().name() << reply.error().message();
-                return;
-            }
-            const auto actions = reply.value();
-            for (const RemoteAction &action : actions) {
-                auto a = addAction(action.id, QIcon::fromTheme(action.iconName), action.text);
-                a->setData(action.id);
-                m_actions[service].append(a);
-            }
-        });
+        QDBusPendingReply<RemoteActions> reply = QDBusConnection::sessionBus().call(getActionsMethod);
+        if (!reply.isValid()) {
+            qCDebug(KRUNNER) << "Error requesting actions; calling" << service << " :" << reply.error().name() << reply.error().message();
+        } else {
+            returnedActions.insert(service, reply.value());
+        }
     }
+    return returnedActions;
+}
+
+void DBusRunner::requestConfig()
+{
+    const QString service = *m_matchingServices.constBegin();
+    auto getConfigMethod = QDBusMessage::createMethodCall(service, m_path, QStringLiteral(IFACE_NAME), QStringLiteral("Config"));
+    QDBusPendingReply<QVariantMap> reply = QDBusConnection::sessionBus().asyncCall(getConfigMethod);
+
+    auto watcher = new QDBusPendingCallWatcher(reply);
+    connect(watcher, &QDBusPendingCallWatcher::finished, this, [this, watcher, service]() {
+        watcher->deleteLater();
+        QDBusReply<QVariantMap> reply = *watcher;
+        if (!reply.isValid()) {
+            suspendMatching(false);
+            qCDebug(KRUNNER) << "Error requesting config; calling" << service << " :" << reply.error().name() << reply.error().message();
+            return;
+        }
+        const QVariantMap config = reply.value();
+        for (auto it = config.cbegin(), end = config.cend(); it != end; ++it) {
+            if (it.key() == QLatin1String("MatchRegex")) {
+                QRegularExpression regex(it.value().toString());
+                setMatchRegex(regex);
+            } else if (it.key() == QLatin1String("MinLetterCount")) {
+                setMinLetterCount(it.value().toInt());
+            } else if (it.key() == QLatin1String("TriggerWords")) {
+                setTriggerWords(it.value().toStringList());
+            } else if (it.key() == QLatin1String("Actions")) {
+                const auto remoteActions = it.value().value<RemoteActions>();
+                createQActionsFromRemoteActions(QMap<QString, RemoteActions>{{service, remoteActions}});
+                m_actionsOnceRequested = true;
+                m_actionsForSessionRequested = true;
+            }
+        }
+        suspendMatching(false);
+    });
 }
 
 void DBusRunner::match(Plasma::RunnerContext &context)
@@ -131,19 +204,29 @@ void DBusRunner::match(Plasma::RunnerContext &context)
     {
         QMutexLocker lock(&m_mutex);
         services = m_matchingServices;
+        m_matchWasCalled = true;
+
+        // Request the actions
+        if ((m_requestActionsOnce && !m_actionsOnceRequested) // We only want to fetch the actions once but haven't done so yet
+            || (!m_actionsForSessionRequested)) { // We want to fetch the actions for each match session
+            m_actionsOnceRequested = true;
+            m_actionsForSessionRequested = true;
+            auto actions = requestActions();
+            const auto actionsArg = QArgument<QMap<QString, RemoteActions>>("QMap<QString, RemoteActions>", actions);
+            QMetaObject::invokeMethod(this, "createQActionsFromRemoteActions", actionsArg);
+        }
     }
     // we scope watchers to make sure the lambda that captures context by reference definitely gets disconnected when this function ends
-    QList<QSharedPointer<QDBusPendingCallWatcher>> watchers;
+    std::vector<std::unique_ptr<QDBusPendingCallWatcher>> watchers;
 
-    for (const QString &service : qAsConst(services)) {
+    for (const QString &service : std::as_const(services)) {
         auto matchMethod = QDBusMessage::createMethodCall(service, m_path, QStringLiteral(IFACE_NAME), QStringLiteral("Match"));
         matchMethod.setArguments(QList<QVariant>({context.query()}));
         QDBusPendingReply<RemoteMatches> reply = QDBusConnection::sessionBus().asyncCall(matchMethod);
 
-        auto watcher = new QDBusPendingCallWatcher(reply);
-        watchers << QSharedPointer<QDBusPendingCallWatcher>(watcher);
+        watchers.push_back(std::make_unique<QDBusPendingCallWatcher>(reply));
         connect(
-            watcher,
+            watchers.back().get(),
             &QDBusPendingCallWatcher::finished,
             this,
             [this, service, &context, reply]() {
@@ -156,7 +239,6 @@ void DBusRunner::match(Plasma::RunnerContext &context)
                     Plasma::QueryMatch m(this);
 
                     m.setText(match.text);
-                    m.setId(match.id);
                     m.setIconName(match.iconName);
                     m.setType(match.type);
                     m.setRelevance(match.relevance);
@@ -165,11 +247,14 @@ void DBusRunner::match(Plasma::RunnerContext &context)
                     m.setUrls(QUrl::fromStringList(match.properties.value(QStringLiteral("urls")).toStringList()));
                     m.setMatchCategory(match.properties.value(QStringLiteral("category")).toString());
                     m.setSubtext(match.properties.value(QStringLiteral("subtext")).toString());
-                    if (match.properties.contains(QStringLiteral("actions"))) {
-                        m.setData(QVariantList({service, match.properties.value(QStringLiteral("actions"))}));
-                    } else {
+                    const auto actionsIt = match.properties.find(QStringLiteral("actions"));
+                    if (actionsIt == match.properties.cend()) {
                         m.setData(QVariantList({service}));
+                    } else {
+                        m.setData(QVariantList({service, actionsIt.value().toStringList()}));
                     }
+                    m.setId(match.id);
+                    m.setMultiLine(match.properties.value(QStringLiteral("multiline")).toBool());
 
                     const QVariant iconData = match.properties.value(QStringLiteral("icon-data"));
                     if (iconData.isValid()) {
@@ -196,7 +281,7 @@ void DBusRunner::match(Plasma::RunnerContext &context)
             Qt::DirectConnection); // process reply in the watcher's thread (aka the one running ::match  not the one owning the runner)
     }
     // we're done matching when every service replies
-    for (auto w : qAsConst(watchers)) {
+    for (auto &w : watchers) {
         w->waitForFinished();
     }
 }
@@ -224,7 +309,12 @@ void DBusRunner::run(const Plasma::RunnerContext &context, const Plasma::QueryMa
     Q_UNUSED(context);
 
     QString actionId;
-    const QString matchId = match.id().mid(id().length() + 1); // QueryMatch::setId mangles the match ID with runnerID + '_'. This unmangles it
+    QString matchId;
+    if (m_hasUniqueResults) {
+        matchId = match.id();
+    } else {
+        matchId = match.id().mid(id().length() + 1); // QueryMatch::setId mangles the match ID with runnerID + '_'. This unmangles it
+    }
     const QString service = match.data().toList().constFirst().toString();
 
     if (match.selectedAction()) {
